@@ -52,33 +52,22 @@ static void     gs_watcher_finalize   (GObject        *object);
 
 static void     initialize_server_extensions (GSWatcher *watcher);
 
-static gboolean check_pointer_timer          (GSWatcher      *watcher);
 static void     schedule_wakeup_event        (GSWatcher      *watcher,
                                               int             when);
+static void     schedule_power_wakeup_event  (GSWatcher      *watcher,
+                                              int             when);
 static gboolean watchdog_timer               (GSWatcher      *watcher);
 static gboolean idle_timer                   (GSWatcher      *watcher);
+static gboolean power_timer                  (GSWatcher      *watcher);
 
 #define GS_WATCHER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GS_TYPE_WATCHER, GSWatcherPrivate))
 
-typedef struct _PointerPosition PointerPosition;
-struct _PointerPosition
-{
-        GdkScreen      *screen;
-        int             x;
-        int             y;
-        GdkModifierType mask;
-};
-
 struct GSWatcherPrivate
 {
         /* settings */
         guint           enabled : 1;
         guint           timeout;
-        guint           pointer_timeout;
-
-        guint           use_xidle_extension : 1;
-        guint           use_sgi_saver_extension : 1;
-        guint           use_mit_saver_extension : 1;
+        guint           power_timeout;
 
         guint           notice_timeout;
 
@@ -86,41 +75,34 @@ struct GSWatcherPrivate
         guint            active : 1;
         guint            idle : 1;
         guint            idle_notice : 1;
-
-        PointerPosition *poll_position;
+        guint            power_notice : 1;
 
         GTimer         *idle_timer;
-        GTimer         *jump_timer;
         guint           emergency_lock : 1;
 
         guint           timer_id;
-        guint           check_pointer_timer_id;
+        guint           power_timer_id;
         guint           watchdog_timer_id;
 
-        guint           using_xidle_extension : 1;
-        guint           using_sgi_saver_extension : 1;
         guint           using_mit_saver_extension : 1;
 
 # ifdef HAVE_MIT_SAVER_EXTENSION
         int             mit_saver_ext_event_number;
         int             mit_saver_ext_error_number;
 # endif
-# ifdef HAVE_SGI_SAVER_EXTENSION
-        int            sgi_saver_ext_event_number;
-        int            sgi_saver_ext_error_number;
-# endif
-
 };
 
 enum {
         IDLE_CHANGED,
         IDLE_NOTICE_CHANGED,
+        POWER_NOTICE_CHANGED,
         LAST_SIGNAL
 };
 
 enum {
         PROP_0,
-        PROP_TIMEOUT
+        PROP_TIMEOUT,
+        PROP_POWER_TIMEOUT
 };
 
 static guint         signals [LAST_SIGNAL] = { 0, };
@@ -158,6 +140,20 @@ gs_watcher_set_timeout (GSWatcher  *watcher,
         }
 }
 
+void
+gs_watcher_set_power_timeout (GSWatcher  *watcher,
+                              guint       timeout)
+{
+        g_return_if_fail (GS_IS_WATCHER (watcher));
+
+        if (watcher->priv->power_timeout != timeout) {
+                watcher->priv->power_timeout = timeout;
+
+                /* restart the timers if necessary */
+                gs_watcher_reset (watcher);
+        }
+}
+
 static void
 gs_watcher_set_property (GObject            *object,
                          guint               prop_id,
@@ -172,6 +168,9 @@ gs_watcher_set_property (GObject            *object,
         case PROP_TIMEOUT:
                 gs_watcher_set_timeout (self, g_value_get_uint (value));
                 break;
+        case PROP_POWER_TIMEOUT:
+                gs_watcher_set_power_timeout (self, g_value_get_uint (value));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -192,6 +191,9 @@ gs_watcher_get_property (GObject            *object,
         case PROP_TIMEOUT:
                 g_value_set_uint (value, self->priv->timeout);
                 break;
+        case PROP_POWER_TIMEOUT:
+                g_value_set_uint (value, self->priv->power_timeout);
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -227,6 +229,16 @@ gs_watcher_class_init (GSWatcherClass *klass)
                               gs_marshal_BOOLEAN__BOOLEAN,
                               G_TYPE_BOOLEAN,
                               1, G_TYPE_BOOLEAN);
+        signals [POWER_NOTICE_CHANGED] =
+                g_signal_new ("power_notice_changed",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GSWatcherClass, power_notice_changed),
+                              NULL,
+                              NULL,
+                              gs_marshal_BOOLEAN__BOOLEAN,
+                              G_TYPE_BOOLEAN,
+                              1, G_TYPE_BOOLEAN);
 
         g_object_class_install_property (object_class,
                                          PROP_TIMEOUT,
@@ -237,6 +249,15 @@ gs_watcher_class_init (GSWatcherClass *klass)
                                                             G_MAXUINT,
                                                             600000,
                                                             G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_TIMEOUT,
+                                         g_param_spec_uint ("power-timeout",
+                                                            NULL,
+                                                            NULL,
+                                                            10000,
+                                                            G_MAXUINT,
+                                                            60000,
+                                                            G_PARAM_READWRITE));
 
         g_type_class_add_private (klass, sizeof (GSWatcherPrivate));
 }
@@ -256,7 +277,8 @@ notice_events_inner (Window   window,
         GdkWindow        *gwindow;
 
         gwindow = gdk_window_lookup (window);
-        if (gwindow && (window != GDK_ROOT_WINDOW ())) {
+        if (gwindow != NULL
+            && (window != GDK_ROOT_WINDOW ())) {
                 /* If it's one of ours, don't mess up its event mask. */
                 return;
         }
@@ -265,14 +287,15 @@ notice_events_inner (Window   window,
         status = XQueryTree (GDK_DISPLAY (), window, &root, &parent, &kids, &nkids);
 
         if (status == 0) {
-                if (kids) {
+                if (kids != NULL) {
                         XFree (kids);
                 }
                 return;
         }
 
-        if (window == root)
+        if (window == root) {
                 top = FALSE;
+        }
 
         XGetWindowAttributes (GDK_DISPLAY (), window, &attrs);
 
@@ -290,6 +313,8 @@ notice_events_inner (Window   window,
 
                 /* Select for PropertyNotify events to get user time changes */
                 events = PropertyChangeMask | events;
+
+                events = PointerMotionMask | PointerMotionHintMask | events;
         } else {
                 /* We want to disable all events */
 
@@ -333,8 +358,8 @@ notice_events_inner (Window   window,
                 }
         }
 
-        if (kids) {
-                while (nkids) {
+        if (kids != NULL) {
+                while (nkids > 0) {
                         notice_events_inner (kids [--nkids], enable, top);
                 }
 
@@ -351,7 +376,7 @@ notice_events (Window   window,
 
         notice_events_inner (window, enable, top);
 
-        gdk_flush ();
+        gdk_display_sync (gdk_display_get_default ());
         gdk_error_trap_pop ();
 }
 
@@ -371,14 +396,26 @@ start_notice_events (GSWatcher *watcher,
         notice_events (window, TRUE, is_top);
 }
 
+static void
+remove_power_timer (GSWatcher *watcher)
+{
+        if (watcher->priv->power_timer_id != 0) {
+                g_source_remove (watcher->priv->power_timer_id);
+                watcher->priv->power_timer_id = 0;
+        }
+}
+
+static void
+add_power_timer (GSWatcher *watcher,
+                 glong      timeout)
+{
+        watcher->priv->power_timer_id = g_timeout_add (timeout, (GSourceFunc)power_timer, watcher);
+}
+
 static void
 remove_idle_timer (GSWatcher *watcher)
 {
         if (watcher->priv->timer_id != 0) {
-                gs_debug ("killing idle_timer  (%u, %u)",
-                          watcher->priv->timeout,
-                          watcher->priv->timer_id);
-
                 g_source_remove (watcher->priv->timer_id);
                 watcher->priv->timer_id = 0;
         }
@@ -389,8 +426,6 @@ add_idle_timer (GSWatcher *watcher,
                 glong      timeout)
 {
         watcher->priv->timer_id = g_timeout_add (timeout, (GSourceFunc)idle_timer, watcher);
-
-        gs_debug ("starting idle_timer (%ld, %u)", timeout, watcher->priv->timer_id);
 }
 
 static void
@@ -416,18 +451,38 @@ add_watchdog_timer (GSWatcher *watcher,
 static void
 reset_timers (GSWatcher *watcher)
 {
-
-        if (watcher->priv->using_mit_saver_extension || watcher->priv->using_sgi_saver_extension) {
-                return;
-        }
-
+        remove_power_timer (watcher);
         remove_idle_timer (watcher);
 
         schedule_wakeup_event (watcher, watcher->priv->timeout);
+        schedule_power_wakeup_event (watcher, watcher->priv->power_timeout);
 
         g_timer_start (watcher->priv->idle_timer);
 }
 
+static gboolean
+_gs_watcher_set_session_power_notice (GSWatcher *watcher,
+                                      gboolean   in_effect)
+{
+        gboolean res;
+
+        res = FALSE;
+
+        if (in_effect != watcher->priv->power_notice) {
+
+                g_signal_emit (watcher, signals [POWER_NOTICE_CHANGED], 0, in_effect, &res);
+                if (res) {
+                        gs_debug ("Changing power notice state: %d", in_effect);
+
+                        watcher->priv->power_notice = in_effect;
+                } else {
+                        gs_debug ("Power notice signal not handled: %d", in_effect);
+                }
+        }
+
+        return res;
+}
+
 static gboolean
 _gs_watcher_set_session_idle_notice (GSWatcher *watcher,
                                      gboolean   in_effect)
@@ -482,7 +537,11 @@ _gs_watcher_notice_activity (GSWatcher *watcher)
                 return;
         }
 
-        gs_debug ("Activity detected: resetting timers");
+        /* if a power notice was sent, cancel it */
+        if (watcher->priv->power_notice) {
+                gboolean in_effect = FALSE;
+                _gs_watcher_set_session_power_notice (watcher, in_effect);
+        }
 
         /* if an idle notice was sent, cancel it */
         if (watcher->priv->idle_notice) {
@@ -508,6 +567,27 @@ _gs_watcher_notice_window_created (GSWatcher *watcher,
         start_notice_events (watcher, window);
 }
 
+static gboolean
+query_pointer_timeout (Window window)
+{
+        Window       root;
+        Window       child;
+        int          root_x;
+        int          root_y;
+        int          win_x;
+        int          win_y;
+        unsigned int mask;
+
+        gdk_error_trap_push ();
+        XQueryPointer (GDK_DISPLAY (),
+                       window,
+                       &root, &child, &root_x, &root_y, &win_x, &win_y, &mask);
+        gdk_display_sync (gdk_display_get_default ());
+        gdk_error_trap_pop ();
+
+        return FALSE;
+}
+
 static void
 gs_watcher_xevent (GSWatcher *watcher,
                    GdkXEvent *xevent)
@@ -539,6 +619,15 @@ gs_watcher_xevent (GSWatcher *watcher,
                         _gs_watcher_notice_window_created (watcher,
                                                            window);
                 }
+                break;
+        case MotionNotify:
+                if (ev->xmotion.is_hint) {
+                        /* need to respond to hints so we continue to get events */
+                        g_timeout_add (1000, (GSourceFunc)query_pointer_timeout, GINT_TO_POINTER (ev->xmotion.window));
+                }
+
+                _gs_watcher_notice_activity (watcher);
+
                 break;
         default:
                 break;
@@ -556,79 +645,17 @@ xevent_filter (GdkXEvent *xevent,
         return GDK_FILTER_CONTINUE;
 }
 
-static void
-remove_check_pointer_timer (GSWatcher *watcher)
-{
-        if (watcher->priv->check_pointer_timer_id != 0) {
-                g_source_remove (watcher->priv->check_pointer_timer_id);
-                watcher->priv->check_pointer_timer_id = 0;
-        }
-}
-
-static void
-add_check_pointer_timer (GSWatcher *watcher,
-                         glong      timeout)
-{
-        watcher->priv->check_pointer_timer_id = g_timeout_add (timeout,
-                                                               (GSourceFunc)check_pointer_timer, watcher);
-
-}
-
-static void
-start_pointer_poll (GSWatcher *watcher)
-{
-        /* run once to set baseline */
-        check_pointer_timer (watcher);
-
-        remove_check_pointer_timer (watcher);
-
-        add_check_pointer_timer (watcher, watcher->priv->pointer_timeout);
-}
-
-static void
-_gs_watcher_pointer_position_free (PointerPosition *pos)
-{
-        if (pos == NULL) {
-                return;
-        }
-
-        g_free (pos);
-        pos = NULL;
-}
-
-static void
-_gs_watcher_set_pointer_position (GSWatcher       *watcher,
-                                  PointerPosition *pos)
-{
-        if (watcher->priv->poll_position != NULL) {
-                _gs_watcher_pointer_position_free (watcher->priv->poll_position);
-        }
-
-        watcher->priv->poll_position = pos;
-}
-
-static void
-stop_pointer_poll (GSWatcher *watcher)
-{
-        remove_check_pointer_timer (watcher);
-
-        _gs_watcher_set_pointer_position (watcher, NULL);
-}
-
 static gboolean
 start_idle_watcher (GSWatcher *watcher)
 {
         g_return_val_if_fail (watcher != NULL, FALSE);
         g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
 
-        g_timer_start (watcher->priv->jump_timer);
-        g_timer_start (watcher->priv->idle_timer);
-
-        start_pointer_poll (watcher);
-
         gdk_window_add_filter (NULL, (GdkFilterFunc)xevent_filter, watcher);
         start_notice_events (watcher, DefaultRootWindow (GDK_DISPLAY ()));
 
+        reset_timers (watcher);
+
         watchdog_timer (watcher);
 
         return FALSE;
@@ -640,12 +667,10 @@ stop_idle_watcher (GSWatcher *watcher)
         g_return_val_if_fail (watcher != NULL, FALSE);
         g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
 
-        g_timer_stop (watcher->priv->jump_timer);
         g_timer_stop (watcher->priv->idle_timer);
 
         remove_idle_timer (watcher);
-
-        stop_pointer_poll (watcher);
+        remove_power_timer (watcher);
 
         stop_notice_events (watcher, DefaultRootWindow (GDK_DISPLAY ()));
         gdk_window_remove_filter (NULL, (GdkFilterFunc)xevent_filter, watcher);
@@ -665,10 +690,21 @@ gs_watcher_get_active (GSWatcher *watcher)
         return active;
 }
 
+static void
+_gs_watcher_reset_state (GSWatcher *watcher)
+{
+        watcher->priv->idle = FALSE;
+        watcher->priv->idle_notice = FALSE;
+        watcher->priv->power_notice = FALSE;
+}
+
 static gboolean
 _gs_watcher_set_active_internal (GSWatcher *watcher,
                                  gboolean   active)
 {
+        /* reset state */
+        _gs_watcher_reset_state (watcher);
+
         if (! active) {
                 watcher->priv->active = FALSE;
                 stop_idle_watcher (watcher);
@@ -746,13 +782,11 @@ gs_watcher_init (GSWatcher *watcher)
         watcher->priv->enabled = TRUE;
         watcher->priv->active = FALSE;
         watcher->priv->timeout = 600000;
-        watcher->priv->pointer_timeout = 1000;
 
         /* time before idle signal to send notice signal */
         watcher->priv->notice_timeout = 10000;
 
         watcher->priv->idle_timer = g_timer_new ();
-        watcher->priv->jump_timer = g_timer_new ();
 
         initialize_server_extensions (watcher);
 
@@ -776,12 +810,8 @@ gs_watcher_finalize (GObject *object)
         watcher->priv->active = FALSE;
         stop_idle_watcher (watcher);
 
-        _gs_watcher_pointer_position_free (watcher->priv->poll_position);
-
         g_timer_destroy (watcher->priv->idle_timer);
         watcher->priv->idle_timer = NULL;
-        g_timer_destroy (watcher->priv->jump_timer);
-        watcher->priv->jump_timer = NULL;
 
         G_OBJECT_CLASS (gs_watcher_parent_class)->finalize (object);
 }
@@ -827,7 +857,8 @@ init_mit_saver_extension (void)
                         gdk_error_trap_push ();
 
                         XKillClient (GDK_DISPLAY (), kill_id);
-                        gdk_flush ();
+
+                        gdk_display_sync (gdk_display_get_default ());
                         gdk_error_trap_pop ();
                 }
 
@@ -845,117 +876,21 @@ init_mit_saver_extension (void)
 #endif /* HAVE_MIT_SAVER_EXTENSION */
 
 
-/* SGI SCREEN_SAVER server extension hackery.
- */
-
-#ifdef HAVE_SGI_SAVER_EXTENSION
-
-# include <X11/extensions/XScreenSaver.h>
-
-gboolean
-query_sgi_saver_extension (int *event_number,
-                           int *error_number)
-{
-        return XScreenSaverQueryExtension (GDK_DISPLAY (),
-                                           event_number,
-                                           error_number);
-}
-
-static gboolean
-init_sgi_saver_extension (gboolean is_blank_active)
-{
-        int         i;
-        GdkDisplay *display   = gdk_display_get_default ();
-        int         n_screens = gdk_display_get_n_screens (display);
-
-        if (is_blank_active)
-                /* If you mess with this while the server thinks it's active,
-                   the server crashes. */
-                return FALSE;
-
-        for (i = 0; i < n_screens; i++) {
-                GdkScreen *screen = gdk_display_get_screen (display, i);
-
-                XScreenSaverDisable (GDK_DISPLAY (), XScreenNumberOfScreen (GDK_SCREEN_XSCREEN (screen)));
-                if (! XScreenSaverEnable (GDK_DISPLAY (), XScreenNumberOfScreen (GDK_SCREEN_XSCREEN (screen)))) {
-
-                        g_warning ("%SGI SCREEN_SAVER extension exists, but can't be initialized;\n"
-                                   "perhaps some other screensaver program is already running?");
-
-                        return FALSE;
-                }
-        }
-
-        return TRUE;
-}
-
-#endif /* HAVE_SGI_SAVER_EXTENSION */
-
-
-/* XIDLE server extension hackery.
- */
-
-#ifdef HAVE_XIDLE_EXTENSION
-
-# include <X11/extensions/xidle.h>
-
-gboolean
-query_xidle_extension (void)
-{
-        int event_number;
-        int error_number;
-
-        return XidleQueryExtension (GDK_DISPLAY (), &event_number, &error_number);
-}
-
-#endif /* HAVE_XIDLE_EXTENSION */
-
 /* If any server extensions have been requested, try and initialize them.
    Issue warnings if requests can't be honored.
 */
 static void
 initialize_server_extensions (GSWatcher *watcher)
 {
-        gboolean server_has_xidle_extension     = FALSE;
-        gboolean server_has_sgi_saver_extension = FALSE;
         gboolean server_has_mit_saver_extension = FALSE;
 
-        watcher->priv->using_xidle_extension     = watcher->priv->use_xidle_extension;
-        watcher->priv->using_sgi_saver_extension = watcher->priv->use_sgi_saver_extension;
-        watcher->priv->using_mit_saver_extension = watcher->priv->use_mit_saver_extension;
+        watcher->priv->using_mit_saver_extension = FALSE;
 
-#ifdef HAVE_XIDLE_EXTENSION
-        server_has_xidle_extension     = query_xidle_extension ();
-#endif
-#ifdef HAVE_SGI_SAVER_EXTENSION
-        server_has_sgi_saver_extension = query_sgi_saver_extension (&watcher->priv->sgi_saver_ext_event_number,
-                                                                    &watcher->priv->sgi_saver_ext_error_number);
-#endif
 #ifdef HAVE_MIT_SAVER_EXTENSION
         server_has_mit_saver_extension = query_mit_saver_extension (&watcher->priv->mit_saver_ext_event_number,
                                                                     &watcher->priv->mit_saver_ext_error_number);
 #endif
 
-        if (! server_has_xidle_extension) {
-                watcher->priv->using_xidle_extension = FALSE;
-        } else {
-                if (watcher->priv->using_xidle_extension) {
-                        gs_debug ("Using XIDLE extension.");
-                } else {
-                        gs_debug ("Not using server's XIDLE extension.");
-                }
-        }
-
-        if (! server_has_sgi_saver_extension) {
-                watcher->priv->using_sgi_saver_extension = FALSE;
-        } else {
-                if (watcher->priv->using_sgi_saver_extension) {
-                        gs_debug ("Using SGI SCREEN_SAVER extension.");
-                } else {
-                        gs_debug ("Not using server's SGI SCREEN_SAVER extension.");
-                }
-        }
-
         if (! server_has_mit_saver_extension) {
                 watcher->priv->using_mit_saver_extension = FALSE;
         } else {
@@ -991,17 +926,6 @@ disable_builtin_screensaver (GSWatcher *watcher,
         desired_prefer_blank    = current_prefer_blank;
         desired_allow_exp       = current_allow_exp;
 
-        /* On SGIs, if interval is non-zero, it is the number of seconds after
-           screen saving starts at which the monitor should be powered down.
-           Obviously I don't want that, so set it to 0 (meaning "never".)
-
-           Power saving is disabled if DontPreferBlanking, but in that case,
-           we don't get extension events either.  So we can't turn it off that way.
-
-           Note: if you're running Irix 6.3 (O2), you may find that your monitor is
-           powering down anyway, regardless of the xset settings.  This is fixed by
-           installing SGI patches 2447 and 2537.
-        */
         desired_server_interval = 0;
 
         /* I suspect (but am not sure) that DontAllowExposures might have
@@ -1009,16 +933,11 @@ disable_builtin_screensaver (GSWatcher *watcher,
            on some systems that don't support XDPMS?  Who know... */
         desired_allow_exp = AllowExposures;
 
-        if (watcher->priv->using_mit_saver_extension || watcher->priv->using_sgi_saver_extension) {
+        if (watcher->priv->using_mit_saver_extension) {
 
                 desired_server_timeout = (watcher->priv->timeout / 1000);
 
-                /* The SGI extension won't give us events unless blanking is on.
-                   I think (unsure right now) that the MIT extension is the opposite. */
-                if (watcher->priv->using_sgi_saver_extension)
-                        desired_prefer_blank = PreferBlanking;
-                else
-                        desired_prefer_blank = DontPreferBlanking;
+                desired_prefer_blank = DontPreferBlanking;
         } else {
                 /* When we're not using an extension, set the server-side timeout to 0,
                    so that the server never gets involved with screen blanking, and we
@@ -1052,7 +971,7 @@ disable_builtin_screensaver (GSWatcher *watcher,
         }
 
 
-#if defined(HAVE_MIT_SAVER_EXTENSION) || defined(HAVE_SGI_SAVER_EXTENSION)
+#if defined(HAVE_MIT_SAVER_EXTENSION)
         {
                 static gboolean extension_initted = FALSE;
 
@@ -1061,24 +980,19 @@ disable_builtin_screensaver (GSWatcher *watcher,
                         extension_initted = TRUE;
 
 # ifdef HAVE_MIT_SAVER_EXTENSION
-                        if (watcher->priv->using_mit_saver_extension)
+                        if (watcher->priv->using_mit_saver_extension) {
                                 init_mit_saver_extension ();
-# endif
-
-# ifdef HAVE_SGI_SAVER_EXTENSION
-                        if (watcher->priv->using_sgi_saver_extension) {
-                                gboolean blank_active = FALSE;
-                                watcher->priv->using_sgi_saver_extension = init_sgi_saver_extension (blank_active);
                         }
 # endif
 
                 }
         }
-#endif /* HAVE_MIT_SAVER_EXTENSION || HAVE_SGI_SAVER_EXTENSION */
+#endif /* HAVE_MIT_SAVER_EXTENSION */
 
-        if (unblank_screen)
+        if (unblank_screen) {
                 /* Turn off the server builtin saver if it is now running. */
                 XForceScreenSaver (GDK_DISPLAY (), ScreenSaverReset);
+        }
 }
 
 static void
@@ -1096,6 +1010,7 @@ maybe_send_signal (GSWatcher *watcher)
 
         if (watcher->priv->idle) {
                 /* already idle, do nothing */
+                gs_debug ("Checking for idleness but already idle");
                 return;
         }
 
@@ -1154,14 +1069,48 @@ maybe_send_signal (GSWatcher *watcher)
         }
 }
 
+static gboolean
+power_timer (GSWatcher *watcher)
+{
+        gint64   elapsed;
+
+        gs_debug ("in power timer");
+
+        watcher->priv->power_timer_id = 0;
+
+        if (! watcher->priv->active) {
+                gs_debug ("Checking for power idleness but watcher is inactive");
+                return FALSE;
+        }
+
+        if (watcher->priv->power_notice) {
+                gs_debug ("Power notice already sent");
+                return FALSE;
+        }
+
+        elapsed = 1000 * g_timer_elapsed (watcher->priv->idle_timer, NULL);
+
+        if (elapsed >= watcher->priv->power_timeout) {
+                gboolean in_effect = TRUE;
+                gs_debug ("Setting power notice elapsed: %ld", (long int)elapsed);
+
+                _gs_watcher_set_session_power_notice (watcher, in_effect);
+        } else {
+                guint time_left;
+
+                time_left = watcher->priv->power_timeout - elapsed;
+                gs_debug ("Scheduling power notice in: %u", time_left);
+                schedule_power_wakeup_event (watcher, time_left);
+        }
+
+        return FALSE;
+}
+
 static gboolean
 idle_timer (GSWatcher *watcher)
 {
         gs_debug ("in idle timer");
 
-        /* try one last time */
-        check_pointer_timer (watcher);
-
         watcher->priv->timer_id = 0;
 
         maybe_send_signal (watcher);
@@ -1169,6 +1118,22 @@ idle_timer (GSWatcher *watcher)
         return FALSE;
 }
 
+static void
+schedule_power_wakeup_event (GSWatcher *watcher,
+                             int        when)
+{
+        guint timeout;
+
+        if (watcher->priv->power_timer_id) {
+                gs_debug ("power_timer already running");
+                return;
+        }
+
+        timeout = when;
+
+        add_power_timer (watcher, timeout);
+}
+
 static void
 schedule_wakeup_event (GSWatcher *watcher,
                        int        when)
@@ -1191,138 +1156,6 @@ schedule_wakeup_event (GSWatcher *watcher,
         add_idle_timer (watcher, timeout);
 }
 
-/* An unfortunate situation is this: the saver is not active, because the
-   user has been typing.  The machine is a laptop.  The user closes the lid
-   and suspends it.  The CPU halts.  Some hours later, the user opens the
-   lid.  At this point, Xt's timers will fire, and the screensaver will blank
-   the screen.
-
-   So far so good -- well, not really, but it's the best that we can do,
-   since the OS doesn't send us a signal *before* shutdown -- but if the
-   user had delayed locking (lockTimeout > 0) then we should start off
-   in the locked state, rather than only locking N minutes from when the
-   lid was opened.  Also, eschewing fading is probably a good idea, to
-   clamp down as soon as possible.
-
-   We only do this when we'd be polling the mouse position anyway.
-   This amounts to an assumption that machines with APM support also
-   have /proc/interrupts.
-*/
-static void
-check_for_clock_skew (GSWatcher *watcher)
-{
-        time_t now;
-        gint64 shift;
-        gint64 i;
-        gint64 i_h;
-        gint64 i_m;
-        gint64 i_s;
-
-        now   = time (NULL);
-        shift = g_timer_elapsed (watcher->priv->jump_timer, NULL);
-        i     = shift;
-
-        i_h = i / (60 * 60);
-        i_m = (i / 60) % 60;
-        i_s = i % 60;
-        gs_debug ("checking wall clock for hibernation, changed: %" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT,
-                  i_h, i_m, i_s);
-
-        if (shift > (watcher->priv->timeout / 1000)) {
-                gint64 s_h;
-                gint64 s_m;
-                gint64 s_s;
-
-                s_h = shift / (60 * 60);
-                s_m = (shift / 60) % 60;
-                s_s = shift % 60;
-
-                gs_debug ("wall clock has jumped by %" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT,
-                          s_h, s_m, s_s);
-
-                watcher->priv->emergency_lock = TRUE;
-                maybe_send_signal (watcher);
-                watcher->priv->emergency_lock = FALSE;
-        }
-
-        g_timer_start (watcher->priv->jump_timer);
-}
-
-static PointerPosition *
-_gs_watcher_pointer_position_read (GSWatcher *watcher)
-{
-        GdkDisplay      *display;
-        PointerPosition *pos;
-
-        pos = g_new0 (PointerPosition, 1);
-
-        display = gdk_display_get_default ();
-
-        gdk_display_get_pointer (display,
-                                 &pos->screen,
-                                 &pos->x,
-                                 &pos->y,
-                                 &pos->mask);
-        return pos;
-}
-
-static gboolean
-_gs_watcher_pointer_position_compare (PointerPosition *pos1,
-                                      PointerPosition *pos2)
-{
-        gboolean changed;
-
-        if (! pos1)
-                return TRUE;
-
-        if (! pos2)
-                return TRUE;
-
-        changed = (pos1->x != pos2->x
-                   || pos1->y != pos2->y
-                   || pos1->screen != pos2->screen
-                   || pos1->mask   != pos2->mask);
-
-        return changed;
-}
-
-static void
-_gs_watcher_check_pointer_position (GSWatcher *watcher)
-{
-        PointerPosition *pos;
-        gboolean         changed;
-        gint64           elapsed;
-
-        pos = _gs_watcher_pointer_position_read (watcher);
-
-        changed = _gs_watcher_pointer_position_compare (watcher->priv->poll_position,
-                                                        pos);
-
-        if (changed) {
-                _gs_watcher_set_pointer_position (watcher, pos);
-                _gs_watcher_notice_activity (watcher);
-        } else {
-                _gs_watcher_pointer_position_free (pos);
-        }
-
-        elapsed = g_timer_elapsed (watcher->priv->idle_timer, NULL);
-        gs_debug ("Idle %" G_GINT64_FORMAT " seconds", elapsed);
-
-        check_for_clock_skew (watcher);
-}
-
-/* When we aren't using a server extension, this timer is used to periodically
-   wake up and poll the mouse position, which is possibly more reliable than
-   selecting motion events on every window.
-*/
-static gboolean
-check_pointer_timer (GSWatcher *watcher)
-{
-        _gs_watcher_check_pointer_position (watcher);
-
-        return TRUE;
-}
-
 /* This timer goes off every few minutes, whether the user is idle or not,
    to try and clean up anything that has gone wrong.
 
